/***************************************************************************
 *   This file is part of the Inanna library.                              *
 *                                                                         *
 *   Copyright (C) 1997-2002 Marko Grnroos <magi@iki.fi>                  *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *  This library is free software; you can redistribute it and/or          *
 *  modify it under the terms of the GNU Library General Public            *
 *  License as published by the Free Software Foundation; either           *
 *  version 2 of the License, or (at your option) any later version.       *
 *                                                                         *
 *  This library is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  Library General Public License for more details.                       *
 *                                                                         *
 *  You should have received a copy of the GNU Library General Public      *
 *  License along with this library; see the file COPYING.LIB.  If         *
 *  not, write to the Free Software Foundation, Inc., 59 Temple Place      *
 *  - Suite 330, Boston, MA 02111-1307, USA.                               *
 *                                                                         *
 ***************************************************************************/

#ifndef __INANNA_ANNETWORK_H__
#define __INANNA_ANNETWORK_H__

#include <magic/mobject.h>
#include <magic/mcoord.h>
#include <magic/mtable.h>
#include <magic/mattribute.h>
#include "inanna/learning.h"
#include "inanna/topology.h"
#include "inanna/neuron.h"

// External predeclarations
class NeuronInitializer;
class Equalizer;

// Internal predeclarations
class FreeWeight;
class Connection;
class Neuron;
class ANNetwork;



///////////////////////////////////////////////////////////////////////////////////
// |   |                             ___                       o                 //
// |\  |  ___                   _   /   \        _    |   ___      _    ___      //
// | \ | /   ) |   | |/\  __  |/ \  |      __  |/ \  -+-  ___| | |/ \  /   ) |/\ //
// |  \| |---  |   | |   /  \ |   | |     /  \ |   |  |  (   | | |   | |---  |   //
// |   |  \__   \__! |   \__/ |   | \___/ \__/ |   |   \  \__| | |   |  \__  |   //
///////////////////////////////////////////////////////////////////////////////////


// We are so f*cking free that all this f*cking freedom restricts our
// f*cking lives so f*cking much

/** Neuron container.
 *
 *  This class abstracts the handling of neurons in a neural network.
 **/
class NeuronContainer : public Array<Neuron> {
 	decl_dynamic (NeuronContainer);
  public:
						NeuronContainer		() : mUnits (*this) {}
						NeuronContainer		(int size) : Array<Neuron> (size), mUnits (*this) {}
						~NeuronContainer	();

	//void				makeNeurons		(int size) {make (size);}

	/** Adds a neuron to the container.
	 *
	 *  The container takes the ownership of the neuron and will
	 *  delete it properly when the time comes.
	 **/
	void				add				(Neuron* neuron);

	/** Removes the i:th unit and all connections from it and to it.
	 **/
	void				removeUnit		(int i);

	/** Writes the container in XML. */
	//void				writeXML		(OStream& out) const;

  private:
	void				disconnectAll	();
	
  protected:
	/** Neurons in the container. */
	Array<Neuron>&	mUnits;
};



///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                _   |   | |   |                           |                //
//               / \  |\  | |\  |  ___   |                  |                //
//              /   \ | \ | | \ | /   ) -+- \    /  __  |/\ | /              //
//              |---| |  \| |  \| |---   |   \\//  /  \ |   |/               //
//              |   | |   | |   |  \__    \   VV   \__/ |   | \              //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

/** A generic and extensible neural network.
 **/
class ANNetwork : public NeuronContainer, virtual public Learner, public Attributed {
	decl_dynamic (ANNetwork);
  public:

						ANNetwork		(const char* description=NULL);
						ANNetwork		(const ANNetwork& orig) {copy (orig);}
	virtual				~ANNetwork	();

	/** Builds a non-connected network according to the given layer
	 *  topology (for example: "7-4-2" or "8").
	 *
	 *  See @ref ANNLayering for some more information about the
	 *  description strings.
	 **/
	void				makeUnits		(const char* topology);

	/** Implementation for @ref LearningMapping.
	 *
	 *  @see #makeUnits
	 **/
	virtual void		make			(const char* top) {makeUnits(top);}

	/** Deletes all the neurons in the network. */
	virtual void		empty			();

	/** Implementation for @ref LearningMapping. */
	virtual void		copy			(const ANNetwork& fnet, bool onlyWeights=false);

	/** Builds the network from connection matrix, restricting
	 *  the topology to only feedforward connections.
	 **/
	void				connectFfw		(const PackTable<int>& conns);

	/** Connects all units in subsequent layers to form feedforward
	 *  topology.
	 **/
	void				connectFullFfw	(bool shortcuts);

	/** Connects all units to all units, including self.
	 **/
	void				connectFull		();

	/** Sets the unit template. The object t must be an instance of an
	 *  inheritor of @ref Neuron.
	 *
	 *  NOTE: The template is destructed along with the ANNetwork,
	 *  so do not destroy it yourself.
	 **/
	void				setUnitPrototype	(Neuron* t) {delete mUnitTemplate; mUnitTemplate=t;}
	const Neuron	*	getUnitPrototype	() const {return mUnitTemplate;}

	/** Implementation for @ref AnyNetwork. Copies the other FreeNet unto self.
	 **/
	virtual void		copyFreeNet		(const ANNetwork& orig, bool onlyWeights=false);

	/** Implementation for @ref AnyNetwork. Initializes the weights and
	 *  biases randomly according to the range, or if an initializer
	 *  has been given, using that.
	 **/
	virtual void		init			(double r=0.0);

	/** Resets activations to 0.0. */
	void				reset			();

	/** Transfers signals once for all units. */
	virtual void		update	 ();

	/** Implementation for LearningMapping. */
	virtual Vector		testPattern		(const PatternSource& set, int pattern) const;

	/** Returns current layering. */
	const ANNTopology&	getTopology		() const {return *mTopology;}
	//const ANNLayering&	getLayering		() const {return *mTopology;}

	/** Sets the given initializer. NOTE: Takes the ownership of the
	 *  initializer.
	 **/
	void				setInitializer	(NeuronInitializer* initer);	

	/** Returns pointer (or NULL) to the equalization handler for the
	 *  network. Non-const version.
	 **/
	Equalizer*			getEqualizer	() {return mpEqualizer;}

	/** Returns pointer (or NULL) to the equalization handler for the
	 *  network. Const version.
	 **/
	const Equalizer*	getEqualizer	() const {return mpEqualizer;}

	/** Sets equalization handler for the network. Gives ownership. */
	void				setEqualizer	(Equalizer* eq=NULL);

	// Tools

	/** Creates a connection from neuron i to neuron j. */
	Connection*			connect			(int i, int j);

	/** Implementation for @ref Object. Prints out a formatted
	 *  printout of the network.
	 **/
	virtual OStream&	operator>>		(OStream& out) const;

	/** Removes unnecessary connections and disables lonely units.
	 *
	 *  @param removeDisableds Should disabled units be purged?
	 *  Default: false (not purged).
	 *
	 *  @param removePassthroughs Remove units that have only one
	 *  input and one output connection. Default: false (not purged).
	 **/
	void				cleanup			(bool removeDisableds=false,
										 bool removePassthroughs=false);

	/** Draws a picture of the network in EPS (Encapsulated
	 *  PostScript). If no size is given, the image is drawn using
	 *  default scaling.
	 *
	 *  @return Picture as EPS code.
	 **/
	String				drawEPS			(double xsize=-1, double ysize=-1) const;

	/** Order the coordinates of the @ref Neuron neurons (which do
	 *  have coordinates) in the network in a pretty way in
	 *  two-dimensional space.
	 **/
	void				drawFeedForward ();

	/** Implementation for @ref Object. */
	void				check			() const;
	
  protected:
	/** Layering information. */
	ANNTopology*		mTopology;

	/** Neuron template. */
	Neuron*				mUnitTemplate;

	/** Neuron initializer method. */
	NeuronInitializer*	mInitializer;

	/** Equalization object. */
	Equalizer*			mpEqualizer;

  private:
	/** Used by drawFeedForward() */
	void				orderColumn		(int colStart, int colEnd);

 	void operator= (const ANNetwork& other) {FORBIDDEN} // Prevent
};

// "Man is truly free only when man truly dies" -- Prince

#endif

