/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2001-2004 Marko Grnroos <magi@iki.fi>                  *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __MAGIC_VM_H_
#define __MAGIC_VM_H_

#include <magic/mpararr.h>
#include <magic/mpackarray.h>

/*******************************************************************************
 * Definitions
 ******************************************************************************/
#define VM_STACK_SIZE    100

/* Exception classes */
EXCEPTIONCLASS (program_overrun);
EXCEPTIONCLASS (execution_error);
EXCEPTIONCLASS (stack_underflow);
EXCEPTIONCLASS (stack_overflow);

/*******************************************************************************
 * Return codes
 ******************************************************************************/

/* General errors. */
#define VMERR_INVALID_ARGS        1
#define VMERR_LOAD_FAILED         2

/* Runtime environment errors. */
#define VMERR_RTE_BASE            100
#define VMERR_KEEP_POSITION       (VMERR_RTE_BASE + 1) // Not an error

/* Compiler errors. */
#define VMERR_COMPILER_BASE       200
#define VMERR_COMPILATION_ERRORS  (VMERR_COMPILER_BASE + 1)

/*******************************************************************************
 * Types
 ******************************************************************************/

/** Return type. 0 indicates success, other values are error codes. */
typedef int vmresult;

/** Virtual machine word type. */
typedef unsigned char vmword;

/* Forward declarations. */
class VM;

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//               |   | |   | ----- |                        |               //
//               |   | |\ /|   |   | _       ___   ___      |               //
//               |   | | V |   |   |/ | |/\ /   )  ___|  ---|               //
//                \ /  | | |   |   |  | |   |---  (   | (   |               //
//                 V   |   |   |   |  | |    \__   \__|  ---|               //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

class VMThread {
  public:
				VMThread	(VM&, VMThread* pParent=NULL, int position=0);
	virtual		~VMThread	();
	vmresult	execute		(PackArray<vmword>& code, List<VMThread*>& threads);
	int			pop			();
	void		push		(int value);
	
  private:
	VM& 	    mrVM;                  //< Reference to the VM.
	int			mThreadId;             //< Thread ID of the thread.
	int			mChildCount;           //< Number of child threads.
	VMThread*	mpParent;              //< The parent thread.
	int			mInstruction;          //< The instruction pointer
	int			mStack[VM_STACK_SIZE]; //< Stack.
	int			mStackPtr;             //< Stack size (stack pointer + 1)
};

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                                 |   | |   |                                //
//                                 |   | |\ /|                                //
//                                 |   | | V |                                //
//                                  \ /  | | |                                //
//                                   V   |   |                                //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

class VM {
  public:
						VM			(int memsize=1000);
	vmresult			init		(int memsize=1000);
	vmresult			load		(const String& objfile);
	vmresult			run			();
	PackArray<vmword>&	memory		() {return mMemory;}

  protected:
	List<VMThread*>		mThreads;
	int					mThreadCount; //< Count the number of threads to know when to exit.
	PackArray<vmword>	mMemory;      //< Instructions

	friend class VMThread;
};

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//          |   | |   | |   |           o             |     o                //
//          |   | |\ /| |\  |  ___   |           ___  |       |              //
//          |   | | V | | \ |  ___| -+- | |   | /   ) |     | |---           //
//           \ /  | | | |  \| (   |  |  |  \ /  |---  |     | |   )          //
//            V   |   | |   |  \__|   \ |   V    \__  |____ | |__/           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

/*******************************************************************************
 * Native function type definitions
 ******************************************************************************/
typedef int vmfresult;
typedef vmfresult (*vmnativefunc) (VM& vm, VMThread& thread);

/*******************************************************************************
 * Native function definition
 ******************************************************************************/
struct VMNativeFunc {
	String       name;
	vmnativefunc func;
};

/*******************************************************************************
 * Native function library
 ******************************************************************************/
class VMNativeLib {
  public:
						VMNativeLib ();
	void				registerFunc	(const String& name, vmnativefunc func);
	int					resolve			(const String& name) const;
	const vmnativefunc	getFunc			(int funcid) const {return mNatives[funcid].func;}
	
  private:
	Array<VMNativeFunc> mNatives;
};

extern VMNativeLib vmNativeLib;
#endif
