/***************************************************************************
 *                                                                         *
 *   Copyright (C) 1998-2004 Marko Grnroos <magi@iki.fi>                  *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <magic/mobject.h>
#include <magic/mpackarray.h>
#include <magic/mlist.h>
#include <magic/mtextstream.h>
#include <vm.h>
#include <vmpriv.h>

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//               |   | |   | ----- |                        |               //
//               |   | |\ /|   |   | _       ___   ___      |               //
//               |   | | V |   |   |/ | |/\ /   )  ___|  ---|               //
//                \ /  | | |   |   |  | |   |---  (   | (   |               //
//                 V   |   |   |   |  | |    \__   \__|  ---|               //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

/*******************************************************************************
 * Default constructor
 ******************************************************************************/
VMThread::VMThread (
	VM&       rVM,        /**< Reference to the VM.        */
	VMThread* pParent,    /**< Parent thread.              */
	int       position    /**< Instruction in the program. */)
		: mrVM (rVM)
{
	mpParent     = pParent;
	mInstruction = position;
	mChildCount  = 0;
	mThreadId    = mrVM.mThreadCount++;
	mInstruction = position;
	mStackPtr    = -1;
}

/*******************************************************************************
 * Destructor
 ******************************************************************************/
VMThread::~VMThread ()
{
	if (mpParent)
		mpParent->mChildCount--;
}

/*******************************************************************************
 * Reads 4-byte value from memory
 ******************************************************************************/
inline int codeToInt (PackArray<vmword>& code, int pos)
{
	return (code[pos]<<24) | (code[pos+1]<<16) | (code[pos+2]<<8) | (code[pos+3]);
}

/*******************************************************************************
 * Stores a 32-bit integer value to memory
 ******************************************************************************/
inline void storeInt (PackArray<vmword>& program, int pos, int value)
{
	program[pos  ] = (vmword) ((value>>24) & 0xff);
	program[pos+1] = (vmword) ((value>>16) & 0xff);
	program[pos+2] = (vmword) ((value>>8) & 0xff);
	program[pos+3] = (vmword) (value & 0xff);
}

/*******************************************************************************
 * Executes the VM thread for one instruction.
 *
 * The thread can create new threads by adding them to the given
 * threads list.
 ******************************************************************************/
vmresult VMThread::execute (
	PackArray<vmword>& code,    /**< The program.                    */
	List<VMThread*>&   threads  /**< Current set of threads running. */)
{
	// Check for program overrun. This is a fatal error.
	if (mInstruction >= code.size())
		throw program_overrun (strformat ("Execution of thread %d overrun end of program.", mThreadId));

	// Execute the current instruction
	switch (code[mInstruction]) {
	  case VMBC_BRK: // Break; exit the thread.
		  if (mChildCount == 0)
			  return true;
		  mInstruction--; // Children still running, stay in position
		  break;
		  
	  case VMBC_NOP: // Do nothing
		  break;
		  
	  case VMBC_FORK: // Create a thread
		  threads.add (new VMThread (mrVM, this, pop()));
		  mChildCount++;
		  break;
		  
	  case VMBC_JOIN: // Join all child threads
		  if (mChildCount != 0) // Continue when all children have exited
			  mInstruction--; // Stay in position
		  break;

	  case VMBC_PUSH: // Push 32-bit value to stack
		  push (codeToInt (code, mInstruction+1));
		  mInstruction += 4;
		  break;

	  case VMBC_DROP:
		  if (mStackPtr < 0)
			  throw stack_underflow ("Stack underflow in DROP operation.");
		  mStackPtr--;
		  break;

	  case VMBC_DUP:
		  if (mStackPtr < 0)
			  throw stack_underflow ("Stack empty in DUP operation.");
		  mStack[++mStackPtr] = mStack[mStackPtr-1];
		  break;

	  case VMBC_SWAP:
		  if (mStackPtr >= 1) {
			  int tmp = mStack[mStackPtr];
			  mStack[mStackPtr] = mStack[mStackPtr-1];
			  mStack[mStackPtr-1] = tmp;
		  }
		  break;

	  case VMBC_ROT:
		  if (mStackPtr >= 2) {
			  int tmp = mStack[mStackPtr-2];
			  mStack[mStackPtr-2] = mStack[mStackPtr-1];
			  mStack[mStackPtr-1] = mStack[mStackPtr];
			  mStack[mStackPtr] = tmp;
		  }
		  break;

	  case VMBC_IFZ: // Jump to address in stack if 2. value in stack is zero
		  if (mStackPtr < 1)
			  throw stack_underflow ("Stack underflow in IFZ operation.");
		  if (mStack[mStackPtr-1] == 0)
			  mInstruction = mStack[mStackPtr] - 1;
		  mStackPtr -= 2; // Drop the address and condition value
		  break;

	  case VMBC_NTVC: { // Native function call
		  // Get native function by the ID
		  int          funcId = codeToInt (code, mInstruction+1);
		  vmnativefunc nfunc  = vmNativeLib.getFunc (funcId);
		  
		  // Execute the function
		  int result = nfunc (mrVM, *this);
		  if (result == VMERR_KEEP_POSITION)
			  mInstruction--; // The function requested to keep calling it.
		  else if (result != 0)
			  throw execution_error ("Error during native function call.\n");
		  else
			  mInstruction += 4; // Jump over native function ID
	  } break;

	  case VMBC_DATA: // Jump over the data block
		  mInstruction += 4 + codeToInt (code, mInstruction+1);
		  break;

	  case VMBC_VAL: // Read value from memory address
		  push (codeToInt (code, pop()));
		  break;

	  case VMBC_STO: { // Store value to memory address
		  int addr = pop();
		  storeInt (code, addr, pop());
	  } break;

	  default: // Unrecognized code.
		  throw execution_error (strformat ("Invalid bytecode '%d'.", code[mInstruction]));
	}

	mInstruction ++;
	
	return false; // Do not exit the thread.
}

/*******************************************************************************
 * Pops and returns a value from the thread's stack
 ******************************************************************************/
int VMThread::pop ()
{
	if (mStackPtr < 0)
		throw stack_underflow ("Stack underflow.");
	return mStack[mStackPtr--];
}

/*******************************************************************************
 * Pushes the given value to the thread's stack
 ******************************************************************************/
void VMThread::push (int value)
{
	if (mStackPtr >= VM_STACK_SIZE)
		throw stack_overflow ("Stack overflow.");
	mStack[++mStackPtr] = value;
}

///////////////////////////////////////////////////////////////////////////////
//                                |   | |   |                                //
//                                |   | |\ /|                                //
//                                |   | | V |                                //
//                                 \ /  | | |                                //
//                                  V   |   |                                //
///////////////////////////////////////////////////////////////////////////////

/*******************************************************************************
 * Default constructor
 ******************************************************************************/
VM::VM (int memsize)
{
	mThreadCount = 0;
	init (memsize);
}

/*******************************************************************************
 * Initializes the VM
 *
 * Exits when all threads have exited
 ******************************************************************************/
vmresult VM::init (int memsize)
{
	// Add the main thread.
	mThreads.add (new VMThread (*this));

	// Allocate and initialize memory
	mMemory.make (memsize);
	for (int i=0; i<memsize; i++)
		mMemory[i] = 0;
	
	return 0;
}

/*******************************************************************************
 * Runs the VM
 *
 * Exits when all threads have exited.
 ******************************************************************************/
vmresult VM::run ()
{
	int threadCount;

	// Run the VM as long as there are threads
	do {
		threadCount = 0;

		// Execute each thread for one instruction
		for (ListIter<VMThread*> i (mThreads); !i.exhausted (); i.next (), threadCount++)
			if (i.get()->execute (mMemory, mThreads)) {
				// The thread exited
				delete i.get();
				i.deleteCurrent ();
			}
	} while (threadCount > 0);

	return 0;
}


