/***************************************************************************
 *   Bytecode compiler from a Forth-like stack language.                   *
 *                                                                         *
 *   Copyright (C) 2004 Marko Grnroos <magi@iki.fi>                       *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <magic/mtextstream.h>
#include <magic/mmap.h>

#include <vm.h>
#include <vmpriv.h>
#include <vmcompiler.h>

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                   |   | |   | |                                           //
//                   |   | |\ /| |      ___       ___                        //
//                   |   | | V | |     /   ) \ / /   ) |/\                   //
//                    \ /  | | | |     |---   X  |---  |                     //
//                     V   |   | |____  \__  / \  \__  |                     //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

VMLexer::VMLexer ()
{
	mIntValue    = 0;
	mCurrentLine = 0;
}

enum lexerstate {
	LXST_NONE = 0,
	LXST_INT,
	LXST_NAME,
	LXST_VARREF,
	LXST_STRING,
	LXST_STRING_QUOTE,
	LXST_COMMENT,
	LXST_END
};

/*******************************************************************************
 * Reads a token from input stream
 ******************************************************************************/
vmtoken VMLexer::readToken (TextIStream& in)
{
	mStringValue.empty();
	mIntValue = 0;

	// Run state machine until a token has been collected
	int state = LXST_NONE;
	while (!in.atEnd()) {
		char c;
		in >> c;
		if (in.atEnd())
			break;
		
		if (c == '\n')
			mCurrentLine ++;

		switch (state) {
		  case LXST_NONE:
			  if (isspace(c))
				  /* Pass */;
			  else if (isdigit(c)) {
				  state = LXST_INT;
				  mIntValue = (int)(c-'0');
			  } else if (c == '"') {
				  state = LXST_STRING;
			  } else if (c == '#') {
				  state = LXST_COMMENT;
			  } else if (c == '$') {
				  state = LXST_VARREF;
			  } else { // In all other cases it's a name character
				  state = LXST_NAME;
				  mStringValue += c;
			  }
			  break;
			  
		  case LXST_INT:
			  if (isdigit(c))
				  mIntValue = mIntValue*10 + ((int)(c-'0'));
			  else if (isspace(c))
				  return VMT_INT;
			  else
				  return VMT_ERROR;
			  break;

		  case LXST_NAME:
			  if (isspace(c))
				  return VMT_NAME;
			  else if (c == ':')
				  return VMT_LABEL;
			  else
				  mStringValue += c;
			  break;

		  case LXST_VARREF:
			  if (isspace(c))
				  return VMT_VARREF;
			  else
				  mStringValue += c;
			  break;

		  case LXST_STRING:
			  if (c == '"')
				  return VMT_LITERAL;
			  else if (c == '\\')
				  state = LXST_STRING_QUOTE;
			  else
				  mStringValue += c;
			  break;

		  case LXST_STRING_QUOTE:
			  if (c == 'n')
				  mStringValue += '\n';
			  else
				  mStringValue += c;
			  state = LXST_STRING;
			  break;

		  case LXST_COMMENT:
			  if (c == '\n')
				  state = LXST_NONE;
			  break;
			  
		  default:
			  // Should never get here
			  return VMT_ERROR;
		};
	}
		
	return VMT_EOS;
}

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//              |   | |   |  ___                  o |                        //
//              |   | |\ /| /   \             --    |  ___                   //
//              |   | | V | |      __  |/|/| |  ) | | /   ) |/\              //
//               \ /  | | | |     /  \ | | | |--  | | |---  |                //
//                V   |   | \___/ \__/ | | | |    | |  \__  |                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

VMCompiler::VMCompiler ()
{
}

/*******************************************************************************
 * Stores value to memory
 ******************************************************************************/
inline void storeInt (String& program, int pos, int value)
{
	program[pos  ] = (char) ((value>>24) & 0xff);
	program[pos+1] = (char) ((value>>16) & 0xff);
	program[pos+2] = (char) ((value>>8) & 0xff);
	program[pos+3] = (char) (value & 0xff);
}

inline void addInt (String& program, int i)
{
	program += "####";
	storeInt (program, program.length()-4, i);
}

/*******************************************************************************
 * Compiles source stream to binary bytecode
 ******************************************************************************/
vmresult VMCompiler::compile (TextIStream& in, String& program)
{
	VMLexer              lexer;
	Map<String,Int>      labels;
	List<CodeReference*> codeReferences;
	int                  errorCount = 0;
	StringMap            variables;

	// Process the source stream.
	// Notice that the parser is currently stateless, that is, it
	// doesn't need access to preceding or following tokens. This
	// creates some limitations for the generated code and dictates some
	// features of the VM. For example, the address of any operation,
	// such as a jump, always has to be stored in stack.
	bool stopParser = false;
	while (!in.atEnd() && !stopParser) {
		// Read a token from the source stream
		vmtoken token = lexer.readToken (in);

		// Generate code for the token
		switch (token) {
		  case VMT_INT:
			  program += (char) VMBC_PUSH;
			  addInt (program, lexer.getInt ());
			  break;
			  
		  case VMT_LITERAL: { // String literal
			  // Store the string in DATA code
			  program += (char) VMBC_DATA;
			  addInt (program, lexer.getString().length()+1); // String length plus \0
			  int strpos = program.length();
			  program += lexer.getString();
			  program += '\0'; // Add terminating zero

			  // Push the string address in stack
			  program += (char) VMBC_PUSH;
			  addInt (program, strpos);
		  } break;

		  case VMT_LABEL:
			  // Store current code position to label storage
			  labels.set (lexer.getString(), new Int ((int) program.length()));
			  break;

		  case VMT_VARREF:
			  // Remember that we must later allocate space for the variable
			  variables.set (lexer.getString(), "1");
			  
			  // This is handled exactly as other references below
			  program += (char)  VMBC_PUSH;
			  codeReferences.add (new CodeReference (program.length(),
													 lexer.getString(),
													 lexer.currentLine()));
			  addInt (program, 0);
			  break;
			  
		  case VMT_NAME: {
			  // A name can be a keyword, native function name, or
			  // label reference.
			  String name = lexer.getString();

			  // Check if the name is a keyword.
			  if (name == "break")
				  program += (char) VMBC_BRK;
			  else if (name == "fork")
				  program += (char) VMBC_FORK;
			  else if (name == "join")
				  program += (char) VMBC_JOIN;
			  else if (name == "dup")
				  program += (char) VMBC_DUP;
			  else if (name == "drop")
				  program += (char) VMBC_DROP;
			  else if (name == "swap")
				  program += (char) VMBC_SWAP;
			  else if (name == "rot")
				  program += (char) VMBC_ROT;
			  else if (name == "if")
				  program += (char) VMBC_IFZ;
			  else if (name == "val")
				  program += (char) VMBC_VAL;
			  else if (name == "sto")
				  program += (char) VMBC_STO;
			  else {
				  // Check if it is a native function call
				  int funcid = vmNativeLib.resolve (lexer.getString());
				  if (funcid != -1) {
					  program += (char) VMBC_NTVC;
					  addInt (program, funcid);
				  } else {
					  // Finally, it might be reference to a label.
					  // The references are resolved later,
					  // so we just store the location now.
					  program += (char)  VMBC_PUSH;
					  codeReferences.add (new CodeReference (program.length(),
															 lexer.getString(),
															 lexer.currentLine()));
					  addInt (program, 0);
				  }
			  }
		  } break;

		  case VMT_ERROR:
			  serr << strformat ("Invalid token at line %d.\n",
								 lexer.currentLine ());
			  stopParser = true;
			  break;
			  
		  case VMT_EOS:
			  break;
		}
	}

	// Add a terminating break, just to be sure
	program += (char) VMBC_BRK;

	// Add the variables into the end of the program
	forStringMap (variables, variter) {
		labels.set (variter.key(), new Int ((int) program.length()));
		addInt (program, 0);
	}

	// Resolve references
	for (ListIter<CodeReference*> coderef_i (codeReferences);
		 !coderef_i.exhausted(); coderef_i.next()) {
		String name = coderef_i.get()->reference;
		if (labels.hasKey(name)) {
			int pos = labels[name].toInt();
			storeInt (program, coderef_i.get()->position, pos);
		} else {
			serr << strformat ("Unresolved label '%s' in line '%d'.\n",
							   (CONSTR) coderef_i.get()->reference,
							   coderef_i.get()->lineno);
			errorCount++;
		}
	}

	if (errorCount > 0) {
		serr << strformat ("%d errors encountered.\n", errorCount);
		return VMERR_COMPILATION_ERRORS;
	}
	return 0;
}

