/***************************************************************************
 *   Command-line application for compiling Forth-like source language     *
 *   to bytecode interpreted by the virtual machine.                       *
 *                                                                         *
 *   Copyright (C) 2004 Marko Grnroos <magi@iki.fi>                       *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <magic/mapplic.h>
#include <magic/mtextstream.h>
#include <vm.h>
#include <vmcompiler.h>

/*******************************************************************************
 * Definitions
 ******************************************************************************/

/* Application exit values */
#define VMCMP_EXIT_INVALID_PARAMETERS  1
#define VMCMP_EXIT_FILE_ERROR          2
#define VMCMP_EXIT_COMPILER_ERROR      3

/*******************************************************************************
 * Prints command-line usage message
 ******************************************************************************/
void usage ()
{
	sout << "Usage: mtvmcmp [options] <mtvm_source_file>\n"
		 << "Compiles source file to virtual machine binary executable.\n"
		 << "The source file name must end with '.vm'.\n"
		 << "Options:\n"
		 << "  -h              Display this usage info\n"
		 << "  -o <filename>   Output file name\n";
}

/*******************************************************************************
 * Loads VM executable from file
 ******************************************************************************/
vmresult load (VM& vm, const String& objfile)
{
	// Open file
	File infile (objfile);
	if (!infile.open (IO_Readable)) {
		serr << "Error: Could not open file '" << objfile << "'.\n";
		return VMERR_LOAD_FAILED;
	}

	// Read file contents to a buffer string
	Ref<String> contents = infile.readAll ();
	if (contents.isNull() || contents->length () == 0) {
		serr << "Error: Empty file.\n";
		return VMERR_LOAD_FAILED;
	}

	// An assumption
	ASSERT (sizeof(vmword) == sizeof(char));
	
	// Copy the file to memory
	int memsize = contents->length();
	vm.memory().make (memsize);
	for (int i=0; i<memsize; i++)
		vm.memory()[i] = contents.object()[i];
	
	return 0;
}

/*******************************************************************************
 * Parameter structure for storing parsed application parameters
 ******************************************************************************/
struct VMCmpParams {
	String srcfilename;
	String exefilename;
};

/*******************************************************************************
 * Parses command-line parameters to parameter structure
 ******************************************************************************/
vmresult parse_cmdline (const Array<String>& params, VMCmpParams& vmparams)
{
	// Read options
	int  currpar      = 0;
	bool options_done = false;
	for (; !options_done && currpar < params.size(); currpar++) {
		if (params[currpar] == "-h") {
			usage ();
			return 1;
		} else if (params[currpar] == "-o" && currpar+1 < params.size())
			vmparams.exefilename = params[++currpar];
		else
			break;
	}

	// Read source file name
	if (params.size() - currpar != 1) {
		usage ();
		return 1;
	}
	vmparams.srcfilename = params[currpar];

	if (isempty(vmparams.exefilename))
		vmparams.exefilename = vmparams.srcfilename + ".vmo";

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                            |   |       o                                  //
//                            |\ /|  ___      _                              //
//                            | V |  ___| | |/ \                             //
//                            | | | (   | | |   |                            //
//                            |   |  \__| | |   |                            //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

Main ()
{
	int result = 0;
		
	/* Read command-line parameters */
	VMCmpParams vmparams;
	result = parse_cmdline (params(), vmparams);
	if (result != 0)
		exit (VMCMP_EXIT_INVALID_PARAMETERS);

	// Open input file
	File srcfile (vmparams.srcfilename);
	if (!srcfile.open (IO_Readable)) {
		serr << "Error: Failed to open source file.\n";
		exit (VMCMP_EXIT_FILE_ERROR);
	}
	TextIStream srcstream (srcfile);

	// Create parser
	VMCompiler compiler;

	// The compiled program
	String program;

	// Compile the program
	result = compiler.compile (srcstream, program);
	if (result != 0) {
		serr << "Errors during compilation. Output not written.\n";
		exit (VMCMP_EXIT_COMPILER_ERROR);
	}

	// Open output file
	File exefile (vmparams.exefilename);
	if (!exefile.open (IO_Writable)) {
		serr << "Error: Failed to open source file.\n";
		exit (VMCMP_EXIT_FILE_ERROR);
	}
	TextOStream exestream (exefile);

	// Write the bytecode to output file
	exestream << program;
}

